% Algorithm 1 - Multivariate signal decomposition

% This program is written by Prof. Milos Dakovic, Faculty of Electrical
% Engineering, University of Montenegro, Podgorica, milos@ac.me

% Please cite:
%[1] LJ. Stankovic, D. Mandic, M. Dakovic, and M. Brajovic, "Time-frequency 
%decomposition of multivariate multicomponent signals," Signal Processing, 
%Volume 142, January 2018, Pages 468-479, http://dx.doi.org/10.1016/j.sigpro.2017.08.001
%[2] LJ. Stankovic, M. Brajovic, M. Dakovic, and D. Mandic, 
%"Two-component Bivariate Signal Decomposition Based on Time-Frequency Analysis," 
%22nd International Conference on Digital Signal Processing IEEE DSP 2017, August 23-25, London, United Kingdom

% To run correctly, this program requires the following suplementary functions: 

% Alg2_min.m   -- Algorithm 2 - Minimization procedure
% TFR_funRT.m  -- Calculates time-frequency representation
% TFR_measure.m -- Calculates the concentration measure of time-frequency
% representation
% Alg_prikaz.m -- It is used for visual presentation of the results during the
% algorithm run
% TFR_fun.m -- Calculates time-frequency representation (required by Alg_prikaz.m)


clear, close


rng('default')
%rand('state',1) %random generator is fixed here to produce results from the paper

IK=2; % Multivariate signal Number
Nc=20; % maximal number of components
EvThreshold=1e-6; % all eigenvalues smaller than EvThreshold * maxEigenvalue will be treated as zero
Delta=0.1;
Epsilon=1e-8;
EpsB=0.0001;

WDorSM=1;

Ls=64;
Ncomp=5; % Required number of components (only for S-method)

WD=0;
for IS=1:IK
    % Signal definition
    N=256;
    t=(-N/2:N/2)';
    
%   signal for Example 1:
%   x0=cos(2.5*cos(300*pi*t/N/64)+2*pi*rand).*exp(-(t/128).^2);
%   signal for Example 1a:
%   x0=cos(t.^4/N/64/N/2-8*pi*t.^2/N/64+2*pi*rand).*exp(-(t/128).^2);

%   signal for Example 2:
    x0=(1.0+0.2)*exp(j*12*pi*t.^2/N/25+j*t.^3/N/N+j*2*pi*rand).*exp(-(t/96).^2)+...
        (1.0-0.1)*exp(j*1/8*pi*t+j*t.^4/N/100/N+j*2*pi*rand).*exp(-(t/128).^2);
    
    x0=x0+0*sqrt(2)*(randn(size(x0))+j*randn(size(x0)));
    
    % Oversampling by 2 and zerro padding
    x=x0;
    X=fft(x);
    X=[X(1:N/2+1);zeros(N,1);X(N/2+2:N+1)];
    x=2*ifft(X);
    x=[zeros(N,1);x;zeros(N,1)];
    X=zeros(2*N+1);
    for k=1:2*N+1
        X(:,k)=x(k+(0:2*N)); %.*hanning(2*N+1); % for PWD
    end
    
    if WDorSM==1
        % WD
        ww=X.*conj(flipud(X));
        WW=fft(ww([N+1:2*N+1,1:N],:));
        WW=real(WW([N+2:2*N+1,1:N+1],:));
        WD=WD+WW;
    else
        % S-method
        F=fft([X(N+1:2*N+1,:);zeros(2*N,2*N+1);X(1:N,:)]);
        F=F([2*N+2:4*N+1,1:2*N+1],:);
        ind=2*N+1+(-N:N);%(N+1):(3*N+1);
        S=abs(F(ind,:)).^2;
        SL=F;SD=conj(F);
        for l=1:Ls
            i1=ind+l;i1(i1>4*N+1)=i1(i1>4*N+1)-(4*N+1);
            i2=ind-l;i2(i2<1)=i2(i2<1)+(4*N+1);
            S=S+2*real(SL(i1,:).*SD(i2,:));
        end
        WD=WD+S;
    end
end
WD=WD/IK;

% Decomposition
WDi=ifft(WD([N+2:2*N+1,1:N+1],:));
WDi=WDi([N+2:2*N+1,1:N+1],:);

% Rotation by 45 degrees
R=zeros(N+1);
for n=1:N+1;
    v=N+n;
    k=n;
    for m=1:N+1;
        R(n,m)=WDi(v,k);
        v=v-1;k=k+1;
    end
end

% Eigenvalues and eigenvectors
opt.disp=0;
[Q,D]=eigs(R,Nc,'lm',opt);
D=abs(diag(D));


figure(12)
T2=abs(TFR_funRT(Q(:,1)));
T22=T2;
subplot(121),imagesc(T2);
title('time-frequency of eigenvector one')

T3=abs(TFR_funRT(Q(:,2)));
T33=T3;
subplot(122),imagesc(T3);
title('time-frequency of eigenvector two')

if WDorSM==1
    Pp=find(abs(D)>EvThreshold*max(abs(D)));
else
    Pp=1:Ncomp;
end
P=length(Pp);

Nupdates=1;
iteration=0;
while Nupdates>0
    Mcomp=zeros(1,P);
    for i=1:P
        Mcomp(i)=TFR_measure(Q(:,Pp(i)));
    end
    [~,ipp]=sort(Mcomp);
    Pp=Pp(ipp);
    
    Nupdates=0;
    Betas=zeros(P);
    for ip=1:P
        Alg_prikaz(Q(:,Pp),2,100*(iteration+1)+ip,1);
        i=Pp(ip);
        B=Alg2_min(Q(:,Pp),ip,Delta,Epsilon);
        if any(abs(B([1:ip-1,ip+1:P]))>EpsB)
            Q(:,i)=Q(:,Pp)*B;
            Q(:,i)=Q(:,i)/norm(Q(:,i));
            for kp=ip+1:P
                k=Pp(kp);
                s=Q(:,i)'*Q(:,k);
                if abs(s)>0.1
                    Q(:,k)=1/sqrt(1-abs(s)^2)*(Q(:,k)-s*Q(:,i));
                end
            end
            Nupdates=Nupdates+1;
        end
        Betas(ip,:)=abs(B');
    end
    Betas
    Nupdates
    iteration=iteration+1;
end

Alg_prikaz(Q(:,Pp),1,iteration,1);

